/*  make_fnirt_field.cpp

    Jesper Andersson, FMRIB Image Analysis Group

    Copyright (C) 2021 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include "utils/options.h"
#include "miscmaths/miscmaths.h"
#include "warpfns/warpfns.h"
#include "warpfns/fnirt_file_reader.h"
#include "warpfns/fnirt_file_writer.h"

// COMMAND LINE OPTIONS

using namespace Utilities;

string title="make_fnirt_field\nCopyright(c) 2021, University of Oxford (Jesper Andersson)";
string examples="make_fnirt_field --inwarp=warpvol --outwarp=newwarpvol --std=1";

Option<bool> verbose(string("-v,--verbose"), false, 
		     string("switch on diagnostic messages"), 
		     false, no_argument);
Option<bool> help(string("-h,--help"), false,
		  string("display this message"),
		  false, no_argument);
Option<bool> debug(string("--debug"), false,
		  string("turn on debugging output"),
		  false, no_argument);
Option<string> inwarp(string("-i,--inwarp"), string(""),
		      string("filename for existing fnirt warp file (--cout)"),
		      true, requires_argument);
Option<string> outwarp(string("-o,--outwarp"), string(""),
		       string("filename for output fnirt warp file"),
		       true, requires_argument);
Option<float> stdevp(string("-s,--std"), 1.0,
		     string("Standard deviation (in mm) of coefficents"),
		     true, requires_argument);
Option<float> fwhm(string("-f,--fwhm"), 1.0,
		   string("FWHM (in coefficients) of smoothing of coefficients"),
		   false, requires_argument);
Option<float> jmin(string("--jmin"), 0.1,
			string("minimum acceptable Jacobian value for constraint (default 0.1)"),
			false, requires_argument);
Option<float> jmax(string("--jmax"), 10.0,
			string("maximum acceptable Jacobian value for constraint (default 10.0)"),
			false, requires_argument);

int make_fnirt_field()
{
  if (verbose.value()) std::cout << "Reading input warps" << std::endl << std::flush;
  // Read existing fnirt field
  NEWIMAGE::FnirtFileReader infile(inwarp.value(),NEWIMAGE::UnknownWarps,verbose.value());

  // Get the field as splinefield
  if (verbose.value()) std::cout << "Making splinefield from input" << std::endl << std::flush;
  std::vector<std::shared_ptr<BASISFIELD::splinefield> > spfield(3);
  for (unsigned int i=0; i<3; i++) {
    BASISFIELD::splinefield test(infile.FieldAsSplinefield(i));
    spfield[i] = std::make_shared<BASISFIELD::splinefield>(infile.FieldAsSplinefield(i));
  }
  // Generate random vectors to set as spline coefficients
  if (verbose.value()) std::cout << "Setting coefficients to a random vector" << std::endl << std::flush;
  NEWIMAGE::volume<float> coef(spfield[0]->CoefSz_x(),spfield[0]->CoefSz_y(),spfield[0]->CoefSz_z());
  std::default_random_engine dre;
  std::normal_distribution<double> ndist(0.0,stdevp.value());
  for (unsigned int i=0; i<3; i++) {
    for (auto it=coef.nsfbegin(); it!=coef.nsfend(); ++it) *it = ndist(dre);
    if (fwhm.value() != 0.0) {
      coef = NEWIMAGE::smooth(coef,fwhm.value()/2.355);
    }
    spfield[i]->SetCoef(coef.vec());
  }
  
  // Check that it hasn't folded over itself
  if (verbose.value()) std::cout << "Making sure field is within prescribed Jacobian range" << std::endl;
  std::vector<unsigned int> imsz = infile.FieldSize();
  std::vector<double> vxsz = infile.VoxelSize();
  NEWIMAGE::volume<float> jac(imsz[0],imsz[1],imsz[2]);
  jac.setxdim(vxsz[0]); jac.setydim(vxsz[1]); jac.setzdim(vxsz[2]); 
  NEWIMAGE::deffield2jacobian(*(spfield[0]),*(spfield[1]),*(spfield[2]),jac);
  float maxj = 0.1, minj = 10.0;
  for (auto it=jac.nsfbegin(); it!=jac.nsfend(); ++it) {
    maxj = std::max(maxj,*it);
    minj = std::min(minj,*it);
  }
  if (minj < jmin.value()) {
    std::cout << "Smallest Jacobian = " << minj << ", smaller than the allowed " << jmin.value() << std::endl;
    exit(EXIT_FAILURE);
  }
  if (maxj > jmax.value()) {
    std::cout << "Largest Jacobian = " << maxj << ", larger than the allowed " << jmax.value() << std::endl;
    exit(EXIT_FAILURE);
  }

  // Write out field
  if (verbose.value()) std::cout << "Writing field out as fnirt coefficent file" << std::endl;
  NEWIMAGE::FnirtFileWriter wrt(outwarp.value(),*(spfield[0]),*(spfield[1]),*(spfield[2]));

  return(EXIT_SUCCESS);
}

int main(int argc, char *argv[])
{
  OptionParser options(title, examples);

  try {
    options.add(inwarp);
    options.add(outwarp);
    options.add(stdevp);
    options.add(fwhm);
    options.add(jmin);
    options.add(jmax);
    options.add(debug);
    options.add(verbose);
    options.add(help);
    
    int nparsed = options.parse_command_line(argc, argv);
    if (nparsed < argc) {
      for (; nparsed<argc; nparsed++) {
        cerr << "Unknown input: " << argv[nparsed] << endl;
      }
      exit(EXIT_FAILURE);
    }

    if ( (help.value()) || (!options.check_compulsory_arguments(true)) )
      {
	options.usage();
	exit(EXIT_FAILURE);
      }
  } catch(X_OptionError& e) {
    options.usage();
    cerr << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  } catch(std::exception &e) {
    cerr << e.what() << endl;
  } 

  return make_fnirt_field();
}
